package asmc

import (
	"encoding/binary"
	"fmt"
	"math"
	"strings"

	"git.urbach.dev/cli/q/src/arm"
	"git.urbach.dev/cli/q/src/asm"
)

func (c *compiler) compileARM(x asm.Instruction) {
	switch x.Mnemonic {
	case asm.CALL:
		switch data := x.Data.(type) {
		case *asm.Label:
			position := Address(len(c.code))
			c.append(arm.Call(0))

			pointer := &pointer{
				Position: position,
				OpSize:   0,
				Size:     4,
			}

			pointer.Resolve = func() Address {
				destination, exists := c.codeLabels[data.Name]

				if !exists {
					panic(fmt.Sprintf("unknown jump label %s", data.Name))
				}

				distance := (destination - position) / 4
				return arm.Call(distance)
			}

			c.codePointers = append(c.codePointers, pointer)
		}

	case asm.LABEL:
		c.codeLabels[x.Data.(*asm.Label).Name] = Address(len(c.code))
		c.append(0xa9be7bfd)
		c.append(0x910003fd)

	case asm.LOAD:
		switch operands := x.Data.(type) {
		case *asm.MemoryRegister:
			if operands.Address.OffsetRegister == math.MaxUint8 {
				c.append(arm.LoadRegister(operands.Register, operands.Address.Base, int16(operands.Address.Offset), operands.Address.Length))
			} else {
				// TODO: LoadDynamicRegister
				panic("not implemented")
			}
		}

	case asm.MOVE:
		switch operands := x.Data.(type) {
		case *asm.RegisterRegister:
			c.append(arm.MoveRegisterRegister(operands.Destination, operands.Source))

		case *asm.RegisterNumber:
			c.append(arm.MoveRegisterNumber(operands.Register, operands.Number))

		case *asm.RegisterLabel:
			position := Address(len(c.code))
			c.append(arm.LoadAddress(operands.Register, 0))

			if strings.HasPrefix(operands.Label, "data ") {
				c.dataPointers = append(c.dataPointers, &pointer{
					Position: position,
					OpSize:   0,
					Size:     4,
					Resolve: func() Address {
						destination, exists := c.dataLabels[operands.Label]

						if !exists {
							panic("unknown label")
						}

						destination += c.dataStart - c.codeStart
						distance := destination - position + 8
						return arm.LoadAddress(operands.Register, int(distance))
					},
				})
			} else {
				panic("not implemented")
			}
		}

	case asm.RETURN:
		c.append(0xa8c27bfd)
		c.append(0xd65f03c0)
		c.append(arm.Return())

	case asm.SYSCALL:
		c.append(arm.Syscall())

	default:
		panic("unknown mnemonic: " + x.Mnemonic.String())
	}
}

func (c *compiler) append(code uint32) {
	c.code = binary.LittleEndian.AppendUint32(c.code, code)
}
