package elf

import (
	"bytes"
	"encoding/binary"
	"io"

	"git.urbach.dev/cli/q/src/config"
	"git.urbach.dev/cli/q/src/exe"
)

const HeaderEnd = HeaderSize + ProgramHeaderSize*2

// ELF represents an ELF file.
type ELF struct {
	Header
	CodeHeader     ProgramHeader
	DataHeader     ProgramHeader
	SectionHeaders []SectionHeader
	StringTable    []byte
}

// Write writes the ELF64 format to the given writer.
func Write(writer io.Writer, codeBytes []byte, dataBytes []byte) {
	sections := exe.MakeSections(HeaderEnd, codeBytes, dataBytes)
	code := sections[0]
	data := sections[1]
	arch := cpu()

	elf := &ELF{
		Header: Header{
			Magic:                       [4]byte{0x7F, 'E', 'L', 'F'},
			Class:                       2,
			Endianness:                  LittleEndian,
			Version:                     1,
			OSABI:                       0,
			ABIVersion:                  0,
			Type:                        TypeExecutable,
			Architecture:                arch,
			FileVersion:                 1,
			EntryPointInMemory:          int64(config.BaseAddress + code.MemoryOffset),
			ProgramHeaderOffset:         HeaderSize,
			SectionHeaderOffset:         0,
			Flags:                       0,
			Size:                        HeaderSize,
			ProgramHeaderEntrySize:      ProgramHeaderSize,
			ProgramHeaderEntryCount:     2,
			SectionHeaderEntrySize:      0,
			SectionHeaderEntryCount:     0,
			SectionNameStringTableIndex: 0,
		},
		CodeHeader: ProgramHeader{
			Type:           ProgramTypeLOAD,
			Flags:          ProgramFlagsExecutable | ProgramFlagsReadable,
			Offset:         int64(code.FileOffset),
			VirtualAddress: int64(config.BaseAddress + code.MemoryOffset),
			SizeInFile:     int64(len(code.Bytes)),
			SizeInMemory:   int64(len(code.Bytes)),
			Align:          int64(config.MemoryAlign),
		},
		DataHeader: ProgramHeader{
			Type:           ProgramTypeLOAD,
			Flags:          ProgramFlagsReadable,
			Offset:         int64(data.FileOffset),
			VirtualAddress: int64(config.BaseAddress + data.MemoryOffset),
			SizeInFile:     int64(len(data.Bytes)),
			SizeInMemory:   int64(len(data.Bytes)),
			Align:          int64(config.MemoryAlign),
		},
	}

	if config.Sections {
		elf.AddSections()
	}

	binary.Write(writer, binary.LittleEndian, &elf.Header)
	binary.Write(writer, binary.LittleEndian, &elf.CodeHeader)
	binary.Write(writer, binary.LittleEndian, &elf.DataHeader)
	writer.Write(bytes.Repeat([]byte{0x00}, code.Padding))
	writer.Write(code.Bytes)
	writer.Write(bytes.Repeat([]byte{0x00}, data.Padding))
	writer.Write(data.Bytes)

	if config.Sections {
		writer.Write(elf.StringTable)
		binary.Write(writer, binary.LittleEndian, &elf.SectionHeaders)
	}
}
